/**
 * @file    shellmatta_ymodem.c
 * @brief   ymodem functions of shellmatta
 * @author  Simon Fischer <dev@s-fischer.net>
 */

#include "shellmatta_ymodem.h"
#include "shellmatta_crc.h"
#include "shellmatta_utils.h"
#include <stddef.h>
#include <string.h>

/** @brief symbols needed for ymodem protocol */
typedef enum {
    YMODEM_NULL    = 0x00, /**< NULL-terminator                 */
    YMODEM_SOH     = 0x01, /**< Start of header                 */
    YMODEM_STX     = 0x02, /**< Start of header for 1k packet   */
    YMODEM_EOT     = 0x04, /**< End of transmission             */
    YMODEM_ACK     = 0x06, /**< Acknowledge                     */
    YMODEM_NAK     = 0x15, /**< Negative acknowledge            */
    YMODEM_CA      = 0x18, /**< Cancel                          */
    YMODEM_SPACE   = 0x20, /**< Space                           */
    YMODEM_CRC     = 0x43  /**< 'C' to indicate CRC type        */
} YMODEM_SYMBOLS;

#define YMODEM_PACKET_SIZE          128u    /**< default packet size of ymodem transmission     */
#define YMODEM_PACKET_SIZE_1K       1024u   /**< extended packet size of ymodem transmission    */
#define YMODEM_CRC_SIZE             2u      /**< CRC size of the ymodem packet                  */

/**
 * @brief               forwards the given character to write-function without formatting
 * @param[in]   handle  shellmatta handle of the instance
 * @param[in]   c       character to be sent
*/
static void shellmatta_ymodem_control(shellmatta_handle_t handle, const uint8_t c)
{
    shellmatta_instance_t *inst = (shellmatta_instance_t*)handle;
    SHELLMATTA_WRITE((char*)&c, 1);
}

/**
 * @brief                                           Initialise the ymodem prior to actually receiving data
 * @param[in, out]  handle                          shellmatta instance handle
 * @param[in]       recvBuffer                      pointer to the buffer to save the received payload in
 * @param[in]       recvPacketCallback              pointer to the file size variable
 * @param[in]       recvPacketCallback              pointer to the packet size variable
 * @param[in]       transmissionCompleteCallback    callback functions for the ymodem module
 * @return          todo
 * @note            Disables the tranport layer if inactive or sets it to mandatory if active
*/
shellmatta_retCode_t shellmatta_ymodem_init(shellmatta_handle_t handle,
                                            uint8_t* recvBuffer,
                                            void (*cancelCallback)(void),
                                            void (*recvPacketCallback)(void),
                                            void (*transmissionCompleteCallback)(void))
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;
    shellmatta_instance_t *inst = (shellmatta_instance_t *)handle;
    (void)memset((void *)&inst->ymodem, 0, sizeof(shellmatta_ymodem_t));

    /** -# check parameters for plausibility  */
    if(     (NULL               != inst)
        &&  (SHELLMATTA_MAGIC   == inst->magic))
    {
        /** -# use instances buffer if no buffer is given */
        if(NULL == recvBuffer)
        {
            if(inst->bufferSize <= YMODEM_PACKET_SIZE_1K)
            {
                /** -# return use fault if buffer is too small */
                ret = SHELLMATTA_USE_FAULT;
            }
            else
            {
                inst->ymodem.packet.packetData = (uint8_t*)inst->buffer;
            }
        }

        /** -# store buffer */
        inst->ymodem.packet.packetData = recvBuffer;

        /** -# init callbacks */
        inst->ymodem.cancelCallback = cancelCallback;
        inst->ymodem.recvPacketCallback = recvPacketCallback;
        inst->ymodem.transmissionCompleteCallback = transmissionCompleteCallback;

        inst->ymodem.state = SHELLMATTA_YMODEM_WAIT_FOR_START;

        /* send initial ymodem symbol to start transmission */
        shellmatta_ymodem_control(handle, YMODEM_CRC);
    }
    else
    {
        ret = SHELLMATTA_USE_FAULT;
    }

    return ret;
}

// /**
//  * @brief                   State machine that processes bytewise input during active ymodem transmission
//  * @param[in, out]  handle  shellmatta handle of the instance
//  * @param[in]       byteIn  the byte to be processed
//  * @note                    won't do anything if ymodem module is not initialized
// */
// void shellmatta_ymodem_receive_packet(shellmatta_handle_t handle, uint8_t byteIn)
// {
//     static uint32_t fileSize = 0u;
//     static char fileSizeStr[7u]; /* hopefully no more bytes than a million will ever be transmitted */
//     shellmatta_ymodem_rcv_retcode_t recvRetCode;

//     //recvRetCode = shellmatta_ymodem_receive_byte(handle, byteIn);
//     switch (shellmatta_ymodem_get_state(handle))
//     {
//     case SHELLMATTA_YMODEM_INACTIVE:
//         /* cant do anything if ymodem module was not correctly initialized */
//         return;
//     case SHELLMATTA_YMODEM_WAIT_FOR_START:
//         /* wait for start symbol of a packet */
//         switch (shellmatta_ymodem_current_data_type)
//         {
//         case YMODEM_NONE: /* go here if the header packet is to be expected */
//             if (recvRetCode == SOH_RECEIVED || recvRetCode == STX_RECEIVED)
//             {
//                 /* reset packet counter */
//                 shellmatta_ymodem_packet_counter = 0u;
//                 shellmatta_ymodem_byte_counter = 0u;
//                 shellmatta_ymodem_set_state(handle, SHELLMATTA_YMODEM_RECEIVE_PACKET);
//                 shellmatta_ymodem_current_data_type = YMODEM_HEADER;
//             }
//             break;
        
//         case YMODEM_HEADER: /* go here if the first body packet is to be expected */
//             if (recvRetCode == STX_RECEIVED || recvRetCode == SOH_RECEIVED)
//             {
//                 shellmatta_ymodem_byte_counter = 0u;
//                 shellmatta_ymodem_set_state(handle, SHELLMATTA_YMODEM_RECEIVE_PACKET);
//                 shellmatta_ymodem_current_data_type = YMODEM_BODY;
//             }
//             break;

//         case YMODEM_BODY: /* go here if the data transmission loop is active */
//             if (recvRetCode == SOH_RECEIVED || recvRetCode == STX_RECEIVED)
//             {
//                 shellmatta_ymodem_byte_counter = 0u;
//                 /* stay in body when SOH was received */
//                 shellmatta_ymodem_set_state(handle, SHELLMATTA_YMODEM_RECEIVE_PACKET);
//                 shellmatta_ymodem_current_data_type = YMODEM_BODY;
//             }
//             else /* go here if the end of transmission symbol is received */
//             if (recvRetCode == EOT_RECEIVED)
//             {
//                 shellmatta_ymodem_byte_counter = 0u;
//                 /* answer with ACK */
//                 shellmatta_ymodem_ack(handle);
//                 /* then send 0x43 symbol */
//                 shellmatta_ymodem_control(handle, YMODEM_CRC);
//                 /* go to footer when EOT was received */
//                 shellmatta_ymodem_current_data_type = YMODEM_FOOTER;
//                 /* then wait for SOH */
//             }
//             break;
//         case YMODEM_FOOTER: /* go here if the end of transmission packet is to be expected */
//             /* it _may_ be, that EOT is sent multiple times, repeat previous procedure */
//             if (recvRetCode == EOT_RECEIVED)
//             {
//                 /* answer with ACK */
//                 shellmatta_ymodem_ack(handle);
//                 /* then send 0x43 symbol */
//                 shellmatta_ymodem_control(handle, YMODEM_CRC);
//             }
//             if (recvRetCode == SOH_RECEIVED)
//             {
//                 /* reset packet counter */
//                 shellmatta_ymodem_packet_counter = 0u;
//                 /* reset byte counter to avoid error by one caused by previously received EOT symbol */
//                 shellmatta_ymodem_byte_counter = 0u;
//                 shellmatta_ymodem_set_state(handle, SHELLMATTA_YMODEM_RECEIVE_PACKET);
//             }
//             break;

//         default:
//             break;
//         }
//         break;
//     case SHELLMATTA_YMODEM_RECEIVE_PACKET:
//         /* receiving data */
//         if (recvRetCode == DATA_RECEIVED)
//         {
//             switch (shellmatta_ymodem_current_data_type)
//             {
//             case YMODEM_HEADER:
//                 switch (shellmatta_ymodem_byte_counter)
//                 {
//                 /* first two bytes are packet number and reverse packet number */
//                 case 0:
//                     shellmatta_ymodem_packet.packetNumber = byteIn;
//                     break;
//                 case 1:
//                     shellmatta_ymodem_packet.reversePacketNumber = byteIn;
//                     break;
//                 default:
//                     break;
//                 }
//                 /* after packet numbers, data field begins */
//                 if ((shellmatta_ymodem_byte_counter >= YMODEM_HEADER_DATA_OFFSET) && 
//                     (shellmatta_ymodem_byte_counter < YMODEM_HEADER_CRC_POSITION))
//                 {
//                     shellmatta_ymodem_packet.packetData[shellmatta_ymodem_byte_counter - YMODEM_HEADER_DATA_OFFSET] = byteIn;
//                     /* check for filename */
//                     if (shellmatta_ymodem_byte_counter < fileNameDelimiterPosition)
//                     {
//                         /* find NULL-Terminator in filename */
//                         if (byteIn == YMODEM_NULL)
//                         {
//                             fileNameDelimiterPosition = shellmatta_ymodem_byte_counter + 1;
//                         }
//                     }
//                     /* after filename the filesize begins */
//                     else
//                     if (shellmatta_ymodem_byte_counter >= fileNameDelimiterPosition && 
//                         (shellmatta_ymodem_byte_counter < fileSizeDelimiterPosition))
//                     {
//                         /* find space as delimiter after filesize */
//                         if (byteIn == YMODEM_SPACE)
//                         {
//                             fileSizeDelimiterPosition = shellmatta_ymodem_byte_counter;
//                             /* convert file size string to actual number */
//                             fileSize = utils_shellAsciiToUInt32(fileSizeStr, fileSizeDelimiterPosition - fileNameDelimiterPosition);
//                             shellmatta_ymodem_packet.fileSize = fileSize;
//                         }
//                         /* save characters in string otherwise */
//                         else
//                         {
//                             fileSizeStr[shellmatta_ymodem_byte_counter - fileNameDelimiterPosition] = byteIn;
//                         }
//                     }
//                     else
//                     {
//                         /* some meta data from sender, possibly file date */
//                         /* not needed for now */
//                     }
//                 }
//                 /* after data field, crc begins */
//                 if (shellmatta_ymodem_byte_counter >= YMODEM_HEADER_CRC_POSITION)
//                 {
//                     if (shellmatta_ymodem_byte_counter == YMODEM_HEADER_CRC_POSITION)
//                     {
//                         /* upper byte */
//                         shellmatta_ymodem_packet.packetCrc = byteIn << 8u;
//                     }
//                     if (shellmatta_ymodem_byte_counter == YMODEM_HEADER_CRC_POSITION + 1)
//                     {
//                         /* lower byte */
//                         shellmatta_ymodem_packet.packetCrc |= byteIn;
//                         /* change ymodem state to check packet */
//                         shellmatta_ymodem_check_packet(handle);
//                     }
//                 }
//                 shellmatta_ymodem_byte_counter++;
//                 break;

//             case YMODEM_BODY:
//                 switch (shellmatta_ymodem_byte_counter)
//                 {
//                 case 0:
//                     shellmatta_ymodem_packet.packetNumber = byteIn;
//                     break;
//                 case 1:
//                     shellmatta_ymodem_packet.reversePacketNumber = byteIn;
//                     break;
//                 default:
//                     break;
//                 }

//                 /* after data field, crc begins */
//                 if (shellmatta_ymodem_byte_counter >= yModemPacketSize + YMODEM_HEADER_DATA_OFFSET)
//                 {
//                     if (shellmatta_ymodem_byte_counter == yModemPacketSize + YMODEM_HEADER_DATA_OFFSET)
//                     {
//                         /* upper byte */
//                         shellmatta_ymodem_packet.packetCrc = byteIn << 8u;
//                     }
//                     if (shellmatta_ymodem_byte_counter == yModemPacketSize + YMODEM_HEADER_DATA_OFFSET + 1)
//                     {
//                         /* lower byte */
//                         shellmatta_ymodem_packet.packetCrc |= byteIn;
//                         /* change ymodem state to check packet */
//                         shellmatta_ymodem_check_packet(handle);
//                     }
//                 }
//                 else
//                 if ((shellmatta_ymodem_byte_counter > 1) && 
//                     (shellmatta_ymodem_byte_counter < (yModemPacketSize + YMODEM_HEADER_DATA_OFFSET)))
//                 {
//                     shellmatta_ymodem_packet.packetData[shellmatta_ymodem_byte_counter - YMODEM_HEADER_DATA_OFFSET] = byteIn;
//                 }
//                 shellmatta_ymodem_byte_counter++;
//                 break;

//             case YMODEM_FOOTER:
//                 /* reset packet counter */
//                 shellmatta_ymodem_packet_counter = 0;

//                 switch (shellmatta_ymodem_byte_counter)
//                 {
//                 case 0:
//                     shellmatta_ymodem_packet.packetNumber = byteIn;
//                     break;
//                 case 1:
//                     shellmatta_ymodem_packet.reversePacketNumber = byteIn;
//                     break;
//                 default:
//                     break;
//                 }

//                 /* after data field, crc begins */
//                 if (shellmatta_ymodem_byte_counter >= yModemPacketSize + YMODEM_HEADER_DATA_OFFSET)
//                 {
//                     if (shellmatta_ymodem_byte_counter == yModemPacketSize + YMODEM_HEADER_DATA_OFFSET)
//                     {
//                         /* upper byte */
//                         shellmatta_ymodem_packet.packetCrc = byteIn << 8u;
//                     }
//                     if (shellmatta_ymodem_byte_counter == yModemPacketSize + YMODEM_HEADER_DATA_OFFSET + 1)
//                     {
//                         /* lower byte */
//                         shellmatta_ymodem_packet.packetCrc |= byteIn;
//                         /* change ymodem state to check packet */
//                         shellmatta_ymodem_check_packet(handle);
//                     }
//                 }
//                 else
//                 if ((shellmatta_ymodem_byte_counter > 1) && 
//                     (shellmatta_ymodem_byte_counter < (yModemPacketSize + YMODEM_HEADER_DATA_OFFSET)))
//                 {
//                     /* save data */
//                     shellmatta_ymodem_packet.packetData[shellmatta_ymodem_byte_counter - YMODEM_HEADER_DATA_OFFSET] = byteIn;
//                 }
//                 shellmatta_ymodem_byte_counter++;
//                 break;
            
//             default:
//                 break;
//             }
//         }
//         break;


//     default:
//         break;
//     }
// }

shellmatta_retCode_t processPacket(shellmatta_handle_t *handle)
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;
    shellmatta_instance_t *inst = (shellmatta_instance_t *)handle;

    (void)inst;

    return ret;
}

static shellmatta_retCode_t ymodem_stateMachine(shellmatta_handle_t *handle, uint8_t byte)
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;
    shellmatta_instance_t *inst = (shellmatta_instance_t *)handle;
    uint16_t computedCrc;

    switch(inst->ymodem.state)
    {
        case SHELLMATTA_YMODEM_INACTIVE:
            /** -# skip state machine if inactive */
            break;
        case SHELLMATTA_YMODEM_WAIT_FOR_START:
            /** -# Wait for start character */
            inst->ymodem.byteCounter = 0u;
            switch(byte)
            {
                case YMODEM_SOH:
                    inst->ymodem.packet.size = YMODEM_PACKET_SIZE;
                    inst->ymodem.state = SHELLMATTA_YMODEM_RECEIVE_HEADER;
                    break;
                case YMODEM_STX:
                    inst->ymodem.packet.size = YMODEM_PACKET_SIZE_1K;
                    inst->ymodem.state = SHELLMATTA_YMODEM_RECEIVE_HEADER;
                    break;
                case YMODEM_EOT:
                    if(NULL != inst->ymodem.transmissionCompleteCallback)
                    {
                        inst->ymodem.transmissionCompleteCallback();
                    }
                    /** -# ACK the successful file reception */
                    shellmatta_ymodem_control(handle, YMODEM_ACK);
                    shellmatta_ymodem_control(handle, YMODEM_CRC);
                    inst->ymodem.state = SHELLMATTA_YMODEM_INACTIVE;
                    break;
                default:
                    /** -# ignore unexpected characters on start */
                    break;
            }
            break;
        case SHELLMATTA_YMODEM_RECEIVE_HEADER:
            if(0u == inst->ymodem.byteCounter)
            {
                inst->ymodem.packet.packetNumber = byte;
                inst->ymodem.byteCounter ++;
            }
            else
            {
                if(((0xffu - byte) != inst->ymodem.packet.packetNumber) ||
                  (inst->ymodem.packetCounter % 256u != inst->ymodem.packet.packetNumber))
                {
                    /** -# return error on packet number mismatch or on unexpected packet numbers */
                    inst->ymodem.state = SHELLMATTA_YMODEM_WAIT_FOR_START;
                    ret = SHELLMATTA_ERROR;
                }
                else
                {
                    /** -# start receiving the payload */
                    inst->ymodem.state = SHELLMATTA_YMODEM_RECEIVE_DATA;
                    inst->ymodem.byteCounter = 0u;
                }
            }
            break;
        case SHELLMATTA_YMODEM_RECEIVE_DATA:
            inst->ymodem.packet.packetData[inst->ymodem.byteCounter] = byte;
            inst->ymodem.byteCounter ++;
            
            /** -# load payload until the packet is full */
            if(inst->ymodem.byteCounter >= inst->ymodem.packet.size)
            {
                inst->ymodem.state = SHELLMATTA_YMODEM_RECEIVE_CRC;
                inst->ymodem.byteCounter = 0u;
                inst->ymodem.packet.crc = 0u;
            }

            break;
        case SHELLMATTA_YMODEM_RECEIVE_CRC:
            inst->ymodem.byteCounter ++;
            inst->ymodem.packet.crc |= (uint16_t)byte << (8u * (YMODEM_CRC_SIZE - inst->ymodem.byteCounter));

            if(inst->ymodem.byteCounter >= YMODEM_CRC_SIZE)
            {
                /** -# check CRC */
                computedCrc = crc16Calc((const char*)inst->ymodem.packet.packetData,
                                        inst->ymodem.packet.size);
                if(computedCrc != inst->ymodem.packet.crc)
                {
                    ret = SHELLMATTA_ERROR;
                }
                else
                {
                    ret = processPacket(handle);
                    if(SHELLMATTA_OK == ret)
                    {
                        /** -# pass data to the application using the callback */
                        if(NULL != inst->ymodem.recvPacketCallback)
                        {
                            inst->ymodem.recvPacketCallback();
                        }

                        //todo - wait - or is it enough to block the recvPacketCallback?

                        /** -# ACK the successful packet reception */
                        shellmatta_ymodem_control(handle, YMODEM_ACK);
                        if(0u == inst->ymodem.packetCounter)
                        {
                            /** -# send addional CRC flag after packet 0 */
                            shellmatta_ymodem_control(handle, YMODEM_CRC);
                        }
                        else
                        {
                            /** -# Calculate the total bytes received */
                            inst->ymodem.totalBytesReceived += inst->ymodem.packet.size;
                        }
                        inst->ymodem.packetCounter ++;
                    }
                }
                inst->ymodem.state = SHELLMATTA_YMODEM_WAIT_FOR_START;
            }
            break;
        default:
            /** -# unexpected state - should never happen */
            break;
    }

    return ret;
}


shellmatta_retCode_t shellmatta_ymodem_processByte(shellmatta_handle_t handle, char byte)
{
    shellmatta_instance_t *inst = (shellmatta_instance_t*)handle;
    shellmatta_retCode_t ret;

    /** -# check if session is cancelled -accept only if ymodem is between packets */
    if(('\x03' == byte) && (inst->ymodem.state == SHELLMATTA_YMODEM_WAIT_FOR_START))
    {
        /** -# explicitly reset ymodem with cancel, if it was active */
        shellmatta_ymodem_reset(handle, true);

        utils_terminateInput(inst);

        ret = SHELLMATTA_ERROR;
    }
    else
    {
        ret = ymodem_stateMachine(handle, (uint8_t)byte);
    }

    return ret;
}


shellmatta_retCode_t shellmatta_ymodem_poll(shellmatta_handle_t handle)
{
    shellmatta_retCode_t ret = SHELLMATTA_ERROR;
    shellmatta_instance_t *inst = (shellmatta_instance_t*)handle;

    if (inst->ymodem.state == SHELLMATTA_YMODEM_WAIT_FOR_START)
    {
        /** -# send ymodem symbol to start transmission */
        shellmatta_ymodem_control(handle, YMODEM_CRC);
        ret = SHELLMATTA_OK;
    }

    return ret;
}



// /**
//  * @brief                   Gives a return code depending on symbol received at different states of the ymodem module
//  * @param[in, out]  handle  shellmatta handle of the instance
//  * @param[in]       byteIn  currently processed input byte
// */
// shellmatta_ymodem_rcv_retcode_t shellmatta_ymodem_receive_byte(shellmatta_handle_t handle, uint8_t byteIn)
// {
//     shellmatta_ymodem_rcv_retcode_t ret;
//     if (SHELLMATTA_YMODEM_WAIT_FOR_START == shellmatta_ymodem_get_state(handle))
//     {
//         switch (byteIn)
//         {
//         case YMODEM_SOH:
//             /* start of header -> default packet size */
//             yModemPacketSize = YMODEM_PACKET_SIZE;
//             ret = SOH_RECEIVED;
//             break;
        
//         case YMODEM_STX:
//             /* start of header -> extended packet size */
//             yModemPacketSize = YMODEM_PACKET_SIZE_1K;
//             ret = STX_RECEIVED;
//             break;

//         case YMODEM_EOT:
//             /* end of transmission -> data transmission complete */
//             ret = EOT_RECEIVED;
//             break;
        
//         case YMODEM_CA:
//             /* cancel transmssion -> reset state machine */
//             ret = CA_RECEIVED;
//             break;

//         default:
//             /* no other data should be received in this state */
//             ret = ERROR;
//             break;
//         }
//         /* give packet size back to upper layer */
//         shellmatta_ymodem_packet.packetSize = yModemPacketSize;
//     }
//     else if (SHELLMATTA_YMODEM_RECEIVE_PACKET == shellmatta_ymodem_get_state(handle))
//     {
//         ret = DATA_RECEIVED;
//     }
//     return ret;
// }

// /**
//  * @brief                   Checks packet data for validity and sends ACK or NAK accordingly
//  * @param[in, out]  handle  shellmatta handle of the instance
//  * @note                    Only to be called after full packet is received
//  * @note                    Will call yModemRecvPacketCallback when packet is received correctly
//  * @note                    Will call ymodemTransmissionCompleteCallback when transmission is completed
// */
// void shellmatta_ymodem_check_packet(shellmatta_handle_t handle)
// {
//     uint16_t computedCrc = 0u;
//     if (shellmatta_ymodem_packet.packetNumber != (0xFF - shellmatta_ymodem_packet.reversePacketNumber))
//     {
//         shellmatta_ymodem_nak(handle);
//         return;
//     }
//     /* compare to internal packet counter */
//     if (shellmatta_ymodem_packet.packetNumber != ((uint8_t)shellmatta_ymodem_packet_counter)) /* uint8_t cast to cause overflow */
//     {
//         shellmatta_ymodem_nak(handle);
//         return;
//     }
//     /* compare packet crc only after other checks succeeded */
//     computedCrc = crc16Calc((const char*)shellmatta_ymodem_packet.packetData, yModemPacketSize);
//     //! uncomment to ignore crc check for debug purposes
//     if (shellmatta_ymodem_packet.packetCrc != computedCrc)
//     {
//         shellmatta_ymodem_nak(handle);
//         return;
//     }
//     /* if previous checks passed, packet is acknowledged */
//     (void) computedCrc;
//     shellmatta_ymodem_total_bytes_received += shellmatta_ymodem_byte_counter;
//     shellmatta_ymodem_byte_counter = 0u;
//     shellmatta_ymodem_ack(handle);
//     if (shellmatta_ymodem_current_data_type == YMODEM_HEADER)
//     {
//         /* send additional CRC symbol if the packet was a header packet */
//         shellmatta_ymodem_control(handle, YMODEM_CRC);
//     }

//     /* callback handling */
//     switch (shellmatta_ymodem_current_data_type)
//     {
//     case YMODEM_NONE:
//         /* shouldn't happen */
//         break;
    
//     case YMODEM_HEADER:
//         /* nothing to do after receiving the header packet */
//         break;

//     case YMODEM_BODY:
//         /* call the ymodem receive packet callback */
//         if (NULL != shellmatta_ymodem_callbacks.yModemRecvPacketCallback)
//         {
//             shellmatta_ymodem_callbacks.yModemRecvPacketCallback();
//         }
//         break;

//     case YMODEM_FOOTER:
//         /* call the ymodem transmission complete callback */
//         if (NULL != shellmatta_ymodem_callbacks.ymodemTransmissionCompleteCallback)
//         {
//             shellmatta_ymodem_callbacks.ymodemTransmissionCompleteCallback();
//         }
//         break;
//     default:
//         break;
//     }
// }

/**
 * @brief                   Acknowledge handling. Will send ACK and set ymodem state back to SHELLMATTA_YMODEM_WAIT_FOR_START
 * @param[in, out]  handle  shellmatta handle of the instance
*/
// void shellmatta_ymodem_ack(shellmatta_handle_t handle)
// {
//     shellmatta_ymodem_control(handle, YMODEM_ACK);
//     shellmatta_ymodem_packet_counter++;
//     shellmatta_ymodem_set_state(handle, SHELLMATTA_YMODEM_WAIT_FOR_START);
// }

/**
 * @brief                   Not-Acknowledge handling. Will send NAK and set ymodem state back to SHELLMATTA_YMODEM_WAIT_FOR_START
 * @param[in, out]  handle  shellmatta handle of the instance
*/
// void shellmatta_ymodem_nak(shellmatta_handle_t handle)
// {
//     shellmatta_ymodem_set_state(handle, SHELLMATTA_YMODEM_WAIT_FOR_START);
//     /* set back current data type to prevent false forward stepping */
//     switch (shellmatta_ymodem_current_data_type)
//     {
//     case YMODEM_NONE:
//         /* no handling needed */
//         break;
    
//     case YMODEM_HEADER:
//         shellmatta_ymodem_control(handle, YMODEM_NAK);
//         break;

//     case YMODEM_BODY:
//         /* YMODEM_BODY stays in YMODEM_BODY */
//         shellmatta_ymodem_control(handle, YMODEM_NAK);
//         shellmatta_ymodem_current_data_type = YMODEM_BODY;
//         break;

//     case YMODEM_FOOTER:
//         /* YMODEM_FOOTER as well */
//         shellmatta_ymodem_current_data_type = YMODEM_FOOTER;
//         break;

//     default:
//         break;
//     }
// }


/**
 * @brief                       reset function for the ymodem module
 * @param[in, out]  ymodem      shellmatta ymomdem instance
 * @param[in]       doCancel    flag to execute the cancel-callback
*/
void shellmatta_ymodem_reset(shellmatta_ymodem_t *ymodem, bool doCancel)
{
    /** -# call cancel callback function */
    if (doCancel)
    {
        /* send cancel symbol */
        //todo - write... shellmatta_ymodem_control(handle, YMODEM_CA);
        if (NULL != ymodem->cancelCallback)
        {
            ymodem->cancelCallback();
        }
    }
    /** -# reset instance variables */
    ymodem->state = SHELLMATTA_YMODEM_INACTIVE;
    ymodem->byteCounter = 0u;
    ymodem->packetCounter = 0u;
    ymodem->totalBytesReceived = 0u;
    ymodem->fileName = NULL;
    ymodem->fileSize = 0u;
    (void)memset((void *)&ymodem->packet, 0, sizeof(shellmatta_ymodem_packet_t));
}

/**
 * @brief       Resets the ymodem module
 * @param[in]   doCancel    Set this flag to execute the cancel-callback function within the ymodem-reset function
 * @note        call this function after file transmission is done or cancelled
*/
void shellmatta_ymodem_end(shellmatta_handle_t handle, bool doCancel)
{
    shellmatta_instance_t *inst = (shellmatta_instance_t *)handle;

    shellmatta_ymodem_reset(&inst->ymodem, doCancel);
    /* clear any possibly leftover inputs */
    utils_clearInput((shellmatta_instance_t*)handle);
}
